#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Saxm2s - an xmms2 client, using python and QT
# Copyright (C) 2005  John O'Meara
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import sys, xmmsclient, math, os
from qt import *
from saxm2s_inc import *

if (len(sys.argv) > 1 and sys.argv[1] == "--debug"):
	import pdb
	pdb.set_trace()

def secondsToMinutesandSeconds(sseconds):
	dmin = sseconds/60
	dsec = sseconds%60
	return str(dmin) + ':' + ('00' + str(dsec) )[-2:]
	
class playerUrlDLG(UrlDialog):
	def __init__(self,parent=None):
		UrlDialog.__init__(self,parent)
		self.url=""
		self.ok=False
		self.connect(self.pushButtonAdd, SIGNAL('clicked()'), self.process)
	def process(self):
		self.url=str(self.lineEditURL.displayText())
		self.ok=True

class playerWinDLG(playerWin):
	def __init__(self,parent=None):
		playerWin.__init__(self,parent)
		self.xmms = xmmsclient.XMMS()
		xmmspath = os.getenv('XMMS_PATH')
		try:
			if xmmspath == None:
				self.xmms.connect()
			else:
				self.xmms.connect(xmmspath)
		except IOError:
			QMessageBox.critical(None, "Saxm2s", "Failed to connect to a running xmms2 server. Start the server before starting SAXm2s")
			sys.exit(1)
		self.songid = 0
		self.songtime = 0
		self.songpos = '-1'
		self.curtime = 0
		self.updateSlider = True
		self.polltimer = QTimer(self, "polltimer")
		self.playlistids = [0]
		self.xmms.broadcast_playlist_changed(self.cmdUpdatePlaylist)
		self.xmms.broadcast_medialib_entry_changed(self.hndMediaChange)
		self.xmms.broadcast_playlist_current_pos(self.hndPosChange)
		self.xmms.broadcast_playback_status(self.hndPlaybackStatusChange)

		res=self.xmms.configval_get('output.volume')
		res.wait()
		res=res.value()
		(l,r)=res.split('/')
		l=int(l)
		r=int(r)
		self.sliderVolume.setValue(max(l,r))
		self.loaditem=[] #should be a list of tuples, in the format of (QListViewItem,
		
		self.connect(self.polltimer, SIGNAL('timeout()'), self.checkstatus)
		self.connect(self.pushButtonBack, SIGNAL('clicked()'), self.cmdPrev)
		self.connect(self.pushButtonPlay, SIGNAL('clicked()'), self.cmdPlay)
		self.connect(self.pushButtonPause, SIGNAL('clicked()'), self.cmdPause)
		self.connect(self.pushButtonStop, SIGNAL('clicked()'), self.cmdStop)
		self.connect(self.pushButtonNext, SIGNAL('clicked()'), self.cmdNext)
		self.connect(self.slider1, SIGNAL('sliderPressed()'), self.sliderUpdateOff)
		self.connect(self.slider1, SIGNAL('sliderReleased()'), self.sliderUpdateOn)
		self.connect(self.slider1, SIGNAL('valueChanged(int)'), self.cmdSeek)
		self.connect(self.listViewPlayList, SIGNAL('doubleClicked(QListViewItem*)'), self.cmdJump)
		self.connect(self.listViewPlayList, SIGNAL('spacePressed(QListViewItem*)'), self.cmdJump)
		self.connect(self.pushButtonAdd, SIGNAL('clicked()'), self.cmdAddFile)
		self.connect(self.pushButtonAddU, SIGNAL('clicked()'), self.cmdAddUrl)
		self.connect(self.sliderVolume, SIGNAL('valueChanged(int)'), self.cmdVolume)
		self.connect(self.sliderBalance, SIGNAL('valueChanged(int)'), self.cmdBalance)
		self.cmdUpdatePlaylist()
		
		self.biglayout = QVBoxLayout(self.TabPage)
		sbar = self.statusBar()
		self.slabel = QLabel(self)
		sbar.addWidget(self.slabel,1,True)

		self.hndPosChange() #this should update the info as far as title and slider go
		res=self.xmms.playback_status()
		res.wait()
		self.hndPlaybackStatusChange(res)
		self.listViewPlayList.setSorting(1, True) # sort by song number ascending
	def songmagnitude(self):
		if len(self.playlistids) == 0:
			return 0
		num=int(math.log(len(self.playlistids),10))
		return num
	def songprepend(self, num):
		prependamount = ""
		for i in range(0, num):
			prependamount = prependamount + '0'
		return prependamount
	def updateTime(self, event = None):
		if not event:
			return # I don't know what to do if this is empty
		event.wait()
		res=event.value()
		self.curtime = res/1000
		if self.updateSlider:
			self.slider1.setValue(self.curtime)
		cmin = self.curtime/60
		csec = (self.curtime)%60
		tmin = self.songtime/60
		tsec = self.songtime%60
		self.textLabelTime.setText(str(cmin) + ':' + ('00' + str(csec) )[-2:] + '/' + str(tmin) + ':' + ('00' + str(tsec))[-2:])
	def checkstatus(self):
		speed = 0
		if self.textLabelState.text() == "Playing":
			speed=250 # if we need to change speed, set to this
			self.updateTime(self.xmms.playback_playtime())
		else:
			speed=500 # if we need to change speed, set to this
		#print self.loaditem
		if len(self.loaditem) > 0:
			donow = []
			self.slabel.setText("updating playlist, %i remaining." % len(self.loaditem))

			if len(self.loaditem) > 4:
				donow = self.loaditem[:4]
				self.loaditem = self.loaditem[3:]
				# there is stuff left over, make our delay less
				self.polltimer.stop()
				self.polltimer.start(100, False)
			else:
				donow = self.loaditem
				self.loaditem = []
				# there is nothing left to do after this run, return to normal speed
				self.polltimer.stop()
				self.polltimer.start(speed, False)
				self.slabel.setText("updating playlist, done.")
			for song in donow:
				songinfo = self.xmms.medialib_get_info(song[2])
				songinfo.wait()
				songinfo=songinfo.get_propdict()
				self.setItemsSongData(song[0],songinfo, song[1])
		if self.songid == 0:
			self.textLabelTime.setText("-:--/-:--")
			if self.updateSlider:
				self.slider1.setValue(0)
	def cmdNext(self):
		res=self.xmms.playlist_set_next_rel(1)
		res.wait()
		res=self.xmms.playback_tickle()
		res.wait()
	def cmdPrev(self):
		res=self.xmms.playlist_set_next_rel(-1)
		res.wait()
		res=self.xmms.playback_tickle()
		res.wait()
	def cmdPlay(self):
		res=self.xmms.playback_status() # 0=stop, 1=play, 2=pause
		res.wait()
		if(res.value() == 1):
			res=self.xmms.playback_tickle()
			res.wait()
		else:
			res=self.xmms.playback_start()
			res.wait()
	def cmdPause(self):
		res=self.xmms.playback_status()
		res.wait()
		if(res.value() == 0):
			return
		elif (res.value() == 2):
			res=self.xmms.playback_start()
			res.wait()
		else:
			res=self.xmms.playback_pause()
			res.wait()
	def cmdStop(self):
		self.songtime=0
		res=self.xmms.playback_stop()
		res.wait()
	def cmdJump(self,item):
		self.xmms.playlist_set_next(int(str(item.text(1)))-1)
		if self.updateSlider:
			self.slider1.setValue(0)
		res=self.xmms.playback_tickle()
		res.wait()
		self.cmdPlay()
	def cmdSeek(self, pos):
		if pos == self.curtime: # this was either called because we set it, or the user dropped it where we are.
			return
		if not self.updateSlider: # don't do anything if we aren't supposed to
			return
		ret=self.xmms.playback_seek_ms(pos*1000)
		ret.wait()
	def sliderUpdateOff(self):
		self.updateSlider=False
	def sliderUpdateOn(self):
		self.updateSlider=True
		self.cmdSeek(self.slider1.value())
	def cmdUpdatePlaylist(self, event=None):
		if (event != None):
			event = event.get_dict()
		#if event:
		#	print event.value()

		# XXX: now that there is a nice XMMSResult emmitted with nice data, if it is received
		# XXX:   something other than redoing the whole list should be done.
		# XXX:
		# XXX: The new data format is like this:
		# XXX: {'position': 30, 'type': 0, 'id': 2} # for new entry
		# XXX: {'type': 1} # for shuffle
		# XXX: {'newposition': 0, 'position': 28, 'type': 4, 'id': 11} # for moved entry
		# XXX: {'type': 5} # for sort, or more likely, reverse list
		# XXX: {'position': 52, 'type': 2} # for removed item
		
		#print "Repopulating playlist"
		if event == None or event['type']==1 or event['type']==5: # redo all
			list = self.xmms.playlist_list()
			list.wait()
			self.listViewPlayList.clear()
			counter = 1
			songlist=list.value()
			self.playlistids = songlist
			for song in songlist:
				#print songinfo
				item = QListViewItem(self.listViewPlayList, None)
				self.loaditem.append((item,str(counter),song))
				self.setItemsSongData(item,None, str(counter))
				counter = counter + 1
			#print "playlist repopulated with " + str(counter) + " entries"
		elif event['type']==0: # new entry
			point=str(event['position']+1)
			item = QListViewItem(self.listViewPlayList, None)
			self.loaditem.append((item,point,event['id']))
			self.setItemsSongData(item,None, point)
			self.playlistids.append(event['id'])
			#self.setItemsSongData(item,songinfo, prependamount+str(event['positon'] + 1)[-(num+1):])
		elif event['type']==2: # removed entry
			pos=event['position']
			num=self.songmagnitude()
			prepend=self.songprepend(num)
			pos=(prepend + str(pos + 1))[-(num+1):]
			item = self.listViewPlayList.findItem(pos,1)
			if (item == None):
				print "error! couldn't find song at " + pos + " for removal."
			else:
				self.listViewPlayList.takeItem(item)
				del item # I don't know if this is the correct way to get rid of an item...
				# then I should renumber all items after me, assuming they exist. This could be costly...
		elif event['type']==3: # clear list
			self.listViewPlayList.clear()
			self.playlistids.clear()
		elif event['type']==4: # move an entry
			print "I don't support moving entries in the playlist yet"
			self.cmdUpdatePlaylist() # reload list as fallback
		else:
			print "unknown action requested of the playlist. event:"
			print event
			self.cmdUpdatePlaylist() # reload as best bet
			return
		return # uncomment to prevent crashes on playlist loads
		# now select the current song and label it as current.
		pos=self.xmms.playlist_current_pos()
		pos.wait()
		pos=pos.value()
		if pos == None:
			return
		pos=int(pos) + 1
		num=self.songmagnitude()
		prepend=self.songprepend(num)
		pos=(prepend + str(pos))[-(num+1):]
		item = self.listViewPlayList.findItem(pos,1)
		if (item == None):
			print "error! couldn't find song " + pos
		else:
			item.setText(0,">")
			self.listViewPlayList.ensureItemVisible(item)
			self.songpos=pos
		
	def cmdAddFile(self):
		bob=QFileDialog("","Audio Files (*.ogg *.mp3 *.wav *.flac);;Ogg Vorbis (*.ogg);;All Files (*)")
		bob.setMode(QFileDialog.ExistingFiles)
		bob.exec_loop()
		for i in bob.selectedFiles():
			res=self.xmms.playlist_add("file://" + str(i))
			#print "file:///" + str(i)
			res.wait()
	def cmdAddUrl(self):
		urlwin=playerUrlDLG(parent=self)
		urlwin.show()
		urlwin.exec_loop()
		if urlwin.ok:
			self.xmms.playlist_add(urlwin.url).wait()
	def cmdVolume(self, newvol):
		#calculate the current balance
		res=self.xmms.configval_get('output.volume')
		res.wait()
		res=res.value()
		(l,r)=res.split('/')
		l=int(l)
		r=int(r)
		newl=0
		newr=0
		if l==r:
			newl=newvol
			newr=newvol
		elif l<r:
			perc=l*100/r
			newl=(newvol*perc/100)
			newr=newvol
		else:
			perc=r*100/l
			newl=newvol
			newr=(newvol*perc/100)
		res=self.xmms.configval_set('output.volume', str(newl) + '/' + str(newr))
		res.wait()
	def cmdBalance(self, newbal):
		res=self.xmms.configval_get('output.volume')
		res.wait()
		res=res.value()
		(l,r)=res.split('/')
		l=int(l)
		r=int(r)
		vol=max(l,r)
		if newbal == 50:
			res=self.xmms.configval_set('output.volume', str(vol) + '/' + str(vol))
			self.textLabelBalPerc.setNum(0)
			self.textLabelBalDir.setText("%C")
		elif newbal > 50: # right side is louder
			newbal = (100-newbal)*2 # 50-(newbal-50), invert and get it between 0 and 50
			res=self.xmms.configval_set('output.volume', str(vol*newbal/100) + '/' + str(vol))
			self.textLabelBalPerc.setNum(100-newbal)
			self.textLabelBalDir.setText("%R")
		else:
			newbal = newbal*2
			res=self.xmms.configval_set('output.volume', str(vol) + '/' + str(vol*newbal/100))
			self.textLabelBalPerc.setNum(100-newbal)
			self.textLabelBalDir.setText("%L")
		
	def hndMediaChange(self, event=None):
		changedid = event.value()
#		print "song with medialib id " + str(changedid) + " has had something change"
		songinfo = self.xmms.medialib_get_info(changedid)
		songinfo.wait()
		songinfo=songinfo.get_propdict()
#		print "\tsearching for positions with the song..."
#		print "\t is " + str(changedid) + " in ",
#		print self.playlistids
		occurances = self.playlistids.count(changedid)
		if occurances == 0: # if its not in our playlist, we don't care
#			print "\tnot found"
			return
		index=0
		num=self.songmagnitude()
		prepend=self.songprepend(num)
		for i in range(0,occurances):
			index = self.playlistids.index(changedid,index)
#			print "\t found at " + str(index)
			pos=(prepend + str(index+1))[-(num+1):]
			item = self.listViewPlayList.findItem(pos,1)
			if item == None:
				print "glitch! should have found item at playlist index " + str(index+1) + " but we didn't!"
			else:
				self.setItemsSongData(item,songinfo, pos)
		if self.songid == songinfo['id']:
			self.updateTitleAndBar(songinfo)
	def setItemsSongData(self,item,songdata, id):
		num=self.songmagnitude()
		item.setText(1, (self.songprepend(num) + str(id))[-(num+1):])
		if (songdata == None):
			return
		if songdata.has_key(('server','title')):
			item.setText(2,songdata[('server','title')])
		if songdata.has_key(('server','artist')):
			item.setText(3,songdata[('server','artist')])
		if songdata.has_key(('server','duration')):
			item.setText(4,secondsToMinutesandSeconds(int(songdata[('server','duration')])/1000))
		else:
			item.setText(4,"-:--")
	def hndPosChange(self, event=None): # called when a new entry has started to be played. Update list to show current song, and fix slider pos.
		if not event:
			event=self.xmms.playlist_current_pos()
			event.wait()
		newpos=event.value()
		if newpos == None: # we aren't playing anything. This is only likely to be reached at startup
			self.updateTitleAndBar()
			return
		res=self.xmms.playlist_list()
		res.wait()
		songlist=res.value()
		songid=songlist[newpos]
		if songid != self.songid:
			self.songid = songid
			res=self.xmms.medialib_get_info(self.songid)
			res.wait()
			songdata=res.get_propdict()
			if (songdata[('server','id')] == '0'):
				pass # no current song
			else:
				self.updateTitleAndBar(songdata)
				self.slider1.setMaxValue(self.songtime)
				pos=newpos + 1
				num=self.songmagnitude()
				prepend=self.songprepend(num)
				pos=(prepend + str(pos))[-(num+1):]
				item = self.listViewPlayList.findItem(self.songpos,1)
				if (item != None):
					item.setText(0, "")
				item = self.listViewPlayList.findItem(pos,1)
				if (item == None):
					print "error! couldn't find song " + pos
				else:
					item.setText(0,">")
					self.songpos=pos
					self.listViewPlayList.ensureItemVisible(item)
				info=""
				for i,j in songdata.iteritems():
					info = info + (str(i) + " = " + str(j)) + "\n"
				self.textBrowserMediaInfo.setText(info)
	def updateTitleAndBar(self, songdata=None):
		if not songdata:
			if len(self.playlistids) == 0:
				return
			res=self.xmms.medialib_get_info(self.songid)
			res.wait()
			songdata=res.get_dict()
		if songdata.has_key(('server','title')):
			self.textLabelTitle.setText(songdata[('server','title')])
			self.setCaption("%s - xmms2" % songdata[('server','title')])
		else:
			self.textLabelTitle.setText("???")
			self.setCaption("xmms2")
		if songdata.has_key(('server','artist')):
			self.textLabelArtist.setText(songdata[('server','artist')])
		else:
			self.textLabelArtist.setText("???")
		if songdata.has_key(('server','duration')):
			self.songtime = int(songdata[('server','duration')]) / 1000
			self.slider1.show()
		else:
			self.songtime = 0
			self.slider1.hide()
	def hndPlaybackStatusChange(self, event=None):
		if not event:
			return # what else to do?
		res=event.get_uint()
		statetext=""
		if (res == xmmsclient.PLAYBACK_STATUS_STOP):
			statetext = "Stopped"
			self.textBrowserMediaInfo.setText("")
		elif (res == xmmsclient.PLAYBACK_STATUS_PLAY):
			statetext = "Playing"
		elif (res == xmmsclient.PLAYBACK_STATUS_PAUSE):
			statetext = "Paused"
		else:
			statetext = "<strong>unknown!</strong>"
		self.textLabelState.setText(statetext)
		if (res != xmmsclient.PLAYBACK_STATUS_PLAY): # if its not playing, go slower
			self.polltimer.stop()
			self.polltimer.start(500, False)
			pass
		else:
			self.polltimer.stop()
			self.polltimer.start(250, False)
			pass
		



if __name__== '__main__':
	app = QApplication(sys.argv)
	QObject.connect(app, SIGNAL('lastWindowClosed()'), app, SLOT('quit()'))
	win = playerWinDLG()
	app.setMainWidget(win)
	win.show()
	app.exec_loop()


#>>> temp.get_hashtable()
#{'resolved': '1', 'album': 'http://www.ocremix.org', 'added': '1116714251', 'laststarted': '1116723681', 'title': 'Space Harrier I got the Spacebeam OC ReMix', 'url': 'file:///home/john/media/Space_Harrier_I_Got_the_Spacebeam_OC_ReMix.mp3', 'bitrate': '160', 'artist': 'Star Salzman', 'date': '2004', 'mime': 'audio/mpeg', 'duration': '176128', 'genre': 'Game', 'lmod': '1116151709', 'id': '2'}
